@@ -1,12 +1,13 @@
 use serde::de::DeserializeOwned;
 use serde::Serialize;
 
-use cosmwasm_std::{to_binary, Binary, BlockInfo, Deps, Env, Order, Pair, StdError, StdResult};
+use cosmwasm_std::{to_binary, Binary, BlockInfo, Deps, Env, Order, Record, StdError, StdResult};
 
 use cw0::maybe_addr;
 use cw721::{
-    AllNftInfoResponse, ApprovedForAllResponse, ContractInfoResponse, CustomMsg, Cw721Query,
-    Expiration, NftInfoResponse, NumTokensResponse, OwnerOfResponse, TokensResponse,
+    AllNftInfoResponse, ApprovalResponse, ApprovalsResponse, ContractInfoResponse, CustomMsg,
+    Cw721Query, Expiration, NftInfoResponse, NumTokensResponse, OperatorsResponse, OwnerOfResponse,
+    TokensResponse,
 };
 use cw_storage_plus::Bound;
 
@@ -52,7 +53,8 @@ where
         })
     }
 
-    fn all_approvals(
+    /// operators returns all operators owner given access to
+    fn operators(
         &self,
         deps: Deps,
         env: Env,
@@ -60,7 +62,7 @@ where
         include_expired: bool,
         start_after: Option<String>,
         limit: Option<u32>,
-    ) -> StdResult<ApprovedForAllResponse> {
+    ) -> StdResult<OperatorsResponse> {
         let limit = limit.unwrap_or(DEFAULT_LIMIT).min(MAX_LIMIT) as usize;
         let start_addr = maybe_addr(deps.api, start_after)?;
         let start = start_addr.map(|addr| Bound::exclusive(addr.as_ref()));
@@ -76,7 +78,58 @@ where
             .take(limit)
             .map(parse_approval)
             .collect();
-        Ok(ApprovedForAllResponse { operators: res? })
+        Ok(OperatorsResponse { operators: res? })
+    }
+
+    fn approval(
+        &self,
+        deps: Deps,
+        env: Env,
+        token_id: String,
+        spender: String,
+        include_expired: bool,
+    ) -> StdResult<ApprovalResponse> {
+        let token = self.tokens.load(deps.storage, &token_id)?;
+        let filtered: Vec<_> = token
+            .approvals
+            .into_iter()
+            .filter(|t| t.spender == spender)
+            .filter(|t| include_expired || !t.is_expired(&env.block))
+            .map(|a| cw721::Approval {
+                spender: a.spender.into_string(),
+                expires: a.expires,
+            })
+            .collect();
+
+        if filtered.is_empty() {
+            return Err(StdError::not_found("Approval not found"));
+        }
+        // we expect only one item
+        let approval = filtered[0].clone();
+
+        Ok(ApprovalResponse { approval })
+    }
+
+    /// approvals returns all approvals owner given access to
+    fn approvals(
+        &self,
+        deps: Deps,
+        env: Env,
+        token_id: String,
+        include_expired: bool,
+    ) -> StdResult<ApprovalsResponse> {
+        let token = self.tokens.load(deps.storage, &token_id)?;
+        let approvals: Vec<_> = token
+            .approvals
+            .into_iter()
+            .filter(|t| include_expired || !t.is_expired(&env.block))
+            .map(|a| cw721::Approval {
+                spender: a.spender.into_string(),
+                expires: a.expires,
+            })
+            .collect();
+
+        Ok(ApprovalsResponse { approvals })
     }
 
     fn tokens(
@@ -111,8 +164,7 @@ where
         limit: Option<u32>,
     ) -> StdResult<TokensResponse> {
         let limit = limit.unwrap_or(DEFAULT_LIMIT).min(MAX_LIMIT) as usize;
-        let start_addr = maybe_addr(deps.api, start_after)?;
-        let start = start_addr.map(|addr| Bound::exclusive(addr.as_ref()));
+        let start = start_after.map(Bound::exclusive);
 
         let tokens: StdResult<Vec<String>> = self
             .tokens
@@ -176,12 +228,12 @@ where
                 token_id,
                 include_expired.unwrap_or(false),
             )?),
-            QueryMsg::ApprovedForAll {
+            QueryMsg::AllOperators {
                 owner,
                 include_expired,
                 start_after,
                 limit,
-            } => to_binary(&self.all_approvals(
+            } => to_binary(&self.operators(
                 deps,
                 env,
                 owner,
@@ -198,11 +250,28 @@ where
             QueryMsg::AllTokens { start_after, limit } => {
                 to_binary(&self.all_tokens(deps, start_after, limit)?)
             }
+            QueryMsg::Approval {
+                token_id,
+                spender,
+                include_expired,
+            } => to_binary(&self.approval(
+                deps,
+                env,
+                token_id,
+                spender,
+                include_expired.unwrap_or(false),
+            )?),
+            QueryMsg::Approvals {
+                token_id,
+                include_expired,
+            } => {
+                to_binary(&self.approvals(deps, env, token_id, include_expired.unwrap_or(false))?)
+            }
         }
     }
 }
 
-fn parse_approval(item: StdResult<Pair<Expiration>>) -> StdResult<cw721::Approval> {
+fn parse_approval(item: StdResult<Record<Expiration>>) -> StdResult<cw721::Approval> {
     item.and_then(|(k, expires)| {
         let spender = String::from_utf8(k)?;
         Ok(cw721::Approval { spender, expires })
